-- some vars
local myappwidth, myappheight = application:getContentWidth(), application:getContentHeight()
local matrix = Matrix.new()

-- react physics 3D
r3d = require "reactphysics3d"
local gravityy = -10
local world = r3d.World.new(0, gravityy, 0) -- 0, -9.8, 0
world.staticbodies = {} -- staticbodies list
world.cubes = {} -- cubes list
world.objs = {} -- objs list
world.missiles = {} -- missiles list

-- tiled
local level = loadfile("tiled/r3d_map01.lua")()
local tilew, tilel = level.tilewidth, level.tileheight
local tiledwidth, tiledlength = level.width * tilew, level.height * tilel

-- a fullscreen 3D viewport
local view = D3.View.new(myappwidth, myappheight, 45, 0.1, tiledlength * tiledlength / 2) -- function D3View:init(width, height, fov, near, far)

-- keyboard control
local isup, isdown, isleft, isright, isshoot, isjump = false, false, false, false, false, false

-- a sky
local sky = Pixel.new(0xffffff, 1, myappwidth * 2, myappheight * 2)
sky:setAnchorPoint(0.5, 0.5)
sky:setPosition(myappwidth / 2, myappheight / 2)
sky:setColor(0x00FFFF, 1, 0x00000FF, 1, 90) -- 0x00040FF -- blue gradient background
stage:addChild(sky)

-- some ground
local ground = PLANE_3D(world, { sizex=tiledwidth / 2, sizez=tiledlength / 2, texpath="3d_objs/texs/groundx01.png", texw=64*64, texh=64*64 })
-- some walls
local wallheight = 64
PLANE_3D( world, { posy=wallheight, posz=-tiledlength / 2, sizex=tiledlength/2, sizez=wallheight, rotx=90, texpath="3d_objs/texs/box.png", texw=16*64, texh=8*64 } )
PLANE_3D( world, { posy=wallheight, posz=tiledlength / 2, sizex=tiledlength, sizez=wallheight, rotx=-90, texpath="3d_objs/texs/box.png", texw=16*64, texh=8*64 } )
PLANE_3D( world, { posx=tiledwidth / 2, posy=wallheight, sizex=tiledwidth, sizez=wallheight, rotx=90, rotz=90, texpath="3d_objs/texs/box.png", texw=16*64, texh=8*64 } )
PLANE_3D( world, { posx=-tiledwidth / 2, posy=wallheight, sizex=tiledwidth, sizez=wallheight, rotx=90, rotz=-90, texpath="3d_objs/texs/box.png", texw=16*64, texh=8*64 } )
-- tiled map
local layers = level.layers
for i = 1, #layers do
	local layer = layers[i]
	local layerType = layer.type
	if layerType == "imagelayer" then
		-- process image layer here
	elseif layerType == "objectgroup" then
		local objects = layer.objects
		local index = 1
		for i = 1, #objects do
			local object = objects[i]
			if object.name == "factory" then -- some buildings
--				local buildingobj, buildingbody = obj_flat_factory(
--						world, "3d_objs/buildings", "fabric1.obj",
--						{ posx=object.x - tiledwidth / 2, posy = 10, posz=object.y - tiledlength / 2,
--						rotx=0, roty=0, rotz=0 } )
			elseif object.name == "roadblock" then -- some buildings
				local buildingobj, buildingbody = obj_factory(
						world, "3d_objs/roads", "road_block.obj",
						{ posx=object.x - tiledwidth / 2, posy = 10, posz=object.y - tiledlength / 2,
						rotx=0, roty=0, rotz=0, mass=750 } )
			elseif object.name == "dustbin" then -- some buildings
				local buildingobj, buildingbody = obj_factory(
						world, "3d_objs/roads", "road_dustbin.obj",
						{ posx=object.x - tiledwidth / 2, posy = 10, posz=object.y - tiledlength / 2,
						rotx=0, roty=0, rotz=0, mass=150 } )
			else
				print("unknown object type")
			end
		end
	else
		print("unknown layer type")
	end
end
-- the player
local shipobj, shipbody = PLAYER(world)

-- event listener
function collisionsEventListener()
	for k, v in pairs(world.missiles) do
		for k1, v1 in pairs(world.cubes) do
			if world:testAABBOverlap(v.xbody, v1.xbody) then
				v.isdirty = true
			end
		end
	end
end
world:setEventListener(collisionsEventListener)

-- the scene
local scene = view:getScene() -- a 3D View sprite
stage:addChild(view)
-- add the objects to the scene
for p = 1, #world.staticbodies do scene:addChild(world.staticbodies[p]) end
for k, v in pairs(world.cubes) do scene:addChild(k) end
for k, v in pairs(world.objs) do scene:addChild(k) end
scene:addChild(shipobj)
-- the light
Lighting.setLight(32 * 16, 8 * 64, -32 * 16, 0.2) -- 15, 30, 0, 0.3
Lighting.setLightTarget(0, 0, 0, 32 * 16, 32 * 16) -- 0, 0, 0, 40, 120

-- the game loop
local ax, ay, az = 0, 0, 0 -- to calculate angles
local bx, by, bz = 0, 0, 0
local dx, dy, dz = bx - ax, by - ay, bz - az -- the angles
local camx, camy, camz -- a camera
local shipweight = 3 * shipobj.frontmass + 1 * shipobj.rearmass -- because ship has 3 front objects and 1 rear object
local force = shipweight * (-gravityy <> gravityy)
local torque = shipweight * 8
local missilelimit = 320 -- the missiles max limit before being removed from the world and the scene
stage:addEventListener(Event.ENTER_FRAME, function(e)
	-- lighting
	Lighting.computeShadows(scene)
	-- destroy
	for k, v in pairs(world.missiles) do
		if -(shipobj:getX() - k:getX()) <> (shipobj:getX() - k:getX()) >= missilelimit or -- x=-x<>x -- faster than math.abs
				-(shipobj:getY() - k:getY()) <> (shipobj:getY() - k:getY()) >= missilelimit or -- x=-x<>x -- faster than math.abs
				-(shipobj:getZ() - k:getZ()) <> (shipobj:getZ() - k:getZ()) >= missilelimit or -- x=-x<>x -- faster than math.abs
				v.isdirty then
			world.missiles[k] = nil
			world:destroyBody(v.xbody)
			scene:removeChild(k)
		end
	end
	-- action!
	world:step(e.deltaTime)
	for k, v in pairs(world.cubes) do k:setMatrix(v.xbody:getTransform()) end
	for k, v in pairs(world.objs) do k:setMatrix(v.xbody:getTransform()) end
	-- the player
	matrix = shipbody:getTransform()
	ax, ay, az = matrix:transformPoint(0, 0, 0)
	bx, by, bz = matrix:transformPoint(0, 0, 1) -- 1 because by default the ship is facing the camera
	dx, dy, dz = bx - ax, by - ay, bz - az
	if isup then shipbody:applyForce(dx * force, 0, dz * force) end
	if isdown then shipbody:applyForce(-dx * force, 0, -dz * force) end
	if isleft then shipbody:applyForce(-dx * force, 0, -dz * force) shipbody:applyTorque(0, -torque, 0) end
	if isright then shipbody:applyForce(-dx * force, 0, -dz * force) shipbody:applyTorque(0, torque, 0) end
	if isjump then shipbody:applyForce(0, force * 18, 0) isjump = false end
	if isshoot then shoot() isshoot = false end
	shipobj:setMatrix(matrix)

	-- the missiles
	for k, v in pairs(world.missiles) do
		matrix = v.xbody:getTransform()
		v.xbody:applyForce(v.xdx * 24 * 64, v.xdy * 24 * 64, v.xdz * 24 * 64)
		v.xbody:applyTorque(0, 0, 2)
		k:setPosition(matrix:getPosition())
	end

	-- the camera FPS style
	camx, camy, camz = shipobj:getX(), shipobj:getY(), shipobj:getZ()
	camx += -dx * 16 -- 16
	camy += 4 -- 4
	camz += -dz * 16 -- 16
	view:lookAt(camx, camy, camz, shipobj:getX() + 0, shipobj:getY() + 6, shipobj:getZ() + 0,	0, 1, 0)
end)

-- shoot
function shoot()
	scene:addChild(obj_factory( world, "3d_objs/ship", "missile.obj",
		{ posx=shipobj:getX() + 0, posy=shipobj:getY() + 0, posz=shipobj:getZ() + 0,
		rotx=dx, roty=dy,rotz=dz, ismissile=true, shooter=shipobj, mass=64 } )
	)
end

-- keyboard
function onKeyDown(e)
	if e.keyCode == KeyCode.UP then isup = true end
	if e.keyCode == KeyCode.DOWN then isdown = true end
	if e.keyCode == KeyCode.LEFT then isleft = true end
	if e.keyCode == KeyCode.RIGHT then isright = true end
	if e.keyCode == KeyCode.W then isshoot = true end
	if e.keyCode == KeyCode.X then isjump = true end
end
function onKeyUp(e)
	if e.keyCode == KeyCode.UP then isup = false end
	if e.keyCode == KeyCode.DOWN then isdown = false end
	if e.keyCode == KeyCode.LEFT then isleft = false end
	if e.keyCode == KeyCode.RIGHT then isright = false end
	if e.keyCode == KeyCode.W then isshoot = false end
	if e.keyCode == KeyCode.X then isjump = false end
end
stage:addEventListener(Event.KEY_DOWN, onKeyDown)
stage:addEventListener(Event.KEY_UP, onKeyUp)
